@@ -44,13 +44,25 @@ typedef struct {
 
 
 static PyObject *
-py_an_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
-  py_annoy *self;
+py_an_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
+  py_annoy *self = (py_annoy *)type->tp_alloc(type, 0);
+  if (self == NULL) {
+    return NULL;
+  }
+  const char *metric = NULL;
 
-  self = (py_annoy *)type->tp_alloc(type, 0);
-  if (self != NULL) {
-    self->f = 0;
-    self->ptr = NULL;
+  static char * kwlist[] = {"f", "metric", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|s", kwlist, &self->f, &metric))
+    return NULL;
+  if (!metric || !strcmp(metric, "angular")) {
+   self->ptr = new AnnoyIndex<int32_t, float, Angular, Kiss64Random>(self->f);
+  } else if (!strcmp(metric, "euclidean")) {
+    self->ptr = new AnnoyIndex<int32_t, float, Euclidean, Kiss64Random>(self->f);
+  } else if (!strcmp(metric, "manhattan")) {
+    self->ptr = new AnnoyIndex<int32_t, float, Manhattan, Kiss64Random>(self->f);
+  } else {
+    PyErr_SetString(PyExc_ValueError, "No such metric");
+    return NULL;
   }
 
   return (PyObject *)self;
@@ -58,19 +70,10 @@ py_an_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
 
 
 static int 
-py_an_init(py_annoy *self, PyObject *args, PyObject *kwds) {
-  const char *metric;
-
-  if (!PyArg_ParseTuple(args, "is", &self->f, &metric))
-    return -1;
-  switch(metric[0]) {
-  case 'a':
-    self->ptr = new AnnoyIndex<int32_t, float, Angular, Kiss64Random>(self->f);
-    break;
-  case 'e':
-    self->ptr = new AnnoyIndex<int32_t, float, Euclidean, Kiss64Random>(self->f);
-    break;
-  }
+py_an_init(py_annoy *self, PyObject *args, PyObject *kwargs) {
+  // Seems to be needed for Python 3
+  const char *metric = NULL;
+  PyArg_ParseTuple(args, "i|s", &self->f, &metric);
   return 0;
 }
 
@@ -85,20 +88,21 @@ py_an_dealloc(py_annoy* self) {
 
 
 static PyMemberDef py_annoy_members[] = {
-  {(char*)"_f", T_INT, offsetof(py_annoy, f), 0,
+  {(char*)"f", T_INT, offsetof(py_annoy, f), 0,
    (char*)""},
   {NULL}	/* Sentinel */
 };
 
 
 static PyObject *
-py_an_load(py_annoy *self, PyObject *args) {
+py_an_load(py_annoy *self, PyObject *args, PyObject *kwargs) {
   char* filename;
   bool res = false;
   if (!self->ptr) 
-    return Py_None;
-  if (!PyArg_ParseTuple(args, "s", &filename))
-    return Py_None;
+    Py_RETURN_NONE;
+  static char * kwlist[] = {"fn", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", kwlist, &filename))
+    Py_RETURN_NONE;
 
   res = self->ptr->load(filename);
 
@@ -111,13 +115,14 @@ py_an_load(py_annoy *self, PyObject *args) {
 
 
 static PyObject *
-py_an_save(py_annoy *self, PyObject *args) {
+py_an_save(py_annoy *self, PyObject *args, PyObject *kwargs) {
   char *filename;
   bool res = false;
   if (!self->ptr) 
-    return Py_None;
-  if (!PyArg_ParseTuple(args, "s", &filename))
-    return Py_None;
+    Py_RETURN_NONE;
+  static char * kwlist[] = {"fn", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s", kwlist, &filename))
+    Py_RETURN_NONE;
 
   res = self->ptr->save(filename);
 
@@ -131,15 +136,15 @@ py_an_save(py_annoy *self, PyObject *args) {
 
 PyObject*
 get_nns_to_python(const vector<int32_t>& result, const vector<float>& distances, int include_distances) {
-  PyObject* l = PyList_New(0);
+  PyObject* l = PyList_New(result.size());
   for (size_t i = 0; i < result.size(); i++)
-    PyList_Append(l, PyInt_FromLong(result[i]));
+    PyList_SetItem(l, i, PyInt_FromLong(result[i]));
   if (!include_distances)
     return l;
 
-  PyObject* d = PyList_New(0);
+  PyObject* d = PyList_New(distances.size());
   for (size_t i = 0; i < distances.size(); i++)
-    PyList_Append(d, PyFloat_FromDouble(distances[i]));
+    PyList_SetItem(d, i, PyFloat_FromDouble(distances[i]));
 
   PyObject* t = PyTuple_New(2);
   PyTuple_SetItem(t, 0, l);
@@ -153,13 +158,16 @@ static PyObject*
 py_an_get_nns_by_item(py_annoy *self, PyObject *args) {
   int32_t item, n, search_k=-1, include_distances=0;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "ii|ii", &item, &n, &search_k, &include_distances))
-    return Py_None;
+    Py_RETURN_NONE;
 
   vector<int32_t> result;
   vector<float> distances;
+
+  Py_BEGIN_ALLOW_THREADS;
   self->ptr->get_nns_by_item(item, n, search_k, &result, include_distances ? &distances : NULL);
+  Py_END_ALLOW_THREADS;
 
   return get_nns_to_python(result, distances, include_distances);
 }
@@ -170,9 +178,9 @@ py_an_get_nns_by_vector(py_annoy *self, PyObject *args) {
   PyObject* v;
   int32_t n, search_k=-1, include_distances=0;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "Oi|ii", &v, &n, &search_k, &include_distances))
-    return Py_None;
+    Py_RETURN_NONE;
 
   vector<float> w(self->f);
   for (int z = 0; z < PyList_Size(v) && z < self->f; z++) {
@@ -182,7 +190,10 @@ py_an_get_nns_by_vector(py_annoy *self, PyObject *args) {
 
   vector<int32_t> result;
   vector<float> distances;
+
+  Py_BEGIN_ALLOW_THREADS;
   self->ptr->get_nns_by_vector(&w[0], n, search_k, &result, include_distances ? &distances : NULL);
+  Py_END_ALLOW_THREADS;
 
   return get_nns_to_python(result, distances, include_distances);
 }
@@ -192,15 +203,15 @@ static PyObject*
 py_an_get_item_vector(py_annoy *self, PyObject *args) {
   int32_t item;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "i", &item))
-    return Py_None;
+    Py_RETURN_NONE;
 
-  vector<float> v;
-  self->ptr->get_item(item, &v);
-  PyObject* l = PyList_New(0);
+  vector<float> v(self->f);
+  self->ptr->get_item(item, &v[0]);
+  PyObject* l = PyList_New(self->f);
   for (int z = 0; z < self->f; z++) {
-    PyList_Append(l, PyFloat_FromDouble(v[z]));
+    PyList_SetItem(l, z, PyFloat_FromDouble(v[z]));
   }
 
   return l;
@@ -209,17 +220,17 @@ py_an_get_item_vector(py_annoy *self, PyObject *args) {
 
 static PyObject* 
 py_an_add_item(py_annoy *self, PyObject *args) {
-  vector<float> w;
-
   PyObject* l;
   int32_t item;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "iO", &item, &l))
-    return Py_None;
-  for (int z = 0; z < PyList_Size(l); z++) {
+    Py_RETURN_NONE;
+
+  vector<float> w(self->f, 0.0);
+  for (int z = 0; z < self->f; z++) {
     PyObject *pf = PyList_GetItem(l,z);
-    w.push_back(PyFloat_AsDouble(pf));
+    w[z] = PyFloat_AsDouble(pf);
   }
   self->ptr->add_item(item, &w[0]);
 
@@ -228,23 +239,39 @@ py_an_add_item(py_annoy *self, PyObject *args) {
 
 
 static PyObject *
-py_an_build(py_annoy *self, PyObject *args) {
+py_an_build(py_annoy *self, PyObject *args, PyObject *kwargs) {
   int q;
   if (!self->ptr) 
-    return Py_None;
-  if (!PyArg_ParseTuple(args, "i", &q))
-    return Py_None;
+    Py_RETURN_NONE;
+  static char * kwlist[] = {"n_trees", NULL};
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i", kwlist, &q))
+    Py_RETURN_NONE;
 
+  Py_BEGIN_ALLOW_THREADS;
   self->ptr->build(q);
+  Py_END_ALLOW_THREADS;
 
   Py_RETURN_TRUE;
 }
 
 
 static PyObject *
-py_an_unload(py_annoy *self, PyObject *args) {
+py_an_unbuild(py_annoy *self) {
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
+  
+  Py_BEGIN_ALLOW_THREADS;
+  self->ptr->unbuild();
+  Py_END_ALLOW_THREADS;
+
+  Py_RETURN_TRUE;
+}
+
+
+static PyObject *
+py_an_unload(py_annoy *self) {
+  if (!self->ptr) 
+    Py_RETURN_NONE;
 
   self->ptr->unload();
 
@@ -254,31 +281,24 @@ py_an_unload(py_annoy *self, PyObject *args) {
 
 static PyObject *
 py_an_get_distance(py_annoy *self, PyObject *args) {
-  int32_t i,j;
-  double d=0;
+  int32_t i, j;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "ii", &i, &j))
-    return Py_None;
-
-  d = self->ptr->get_distance(i,j);
+    Py_RETURN_NONE;
 
+  double d = self->ptr->get_distance(i,j);
   return PyFloat_FromDouble(d);
 }
 
 
 static PyObject *
-py_an_get_n_items(py_annoy *self, PyObject *args) {
-  int32_t n=0;
-  bool is_n=false;
+py_an_get_n_items(py_annoy *self) {
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
 
-  n = self->ptr->get_n_items();
-  is_n = true;
-
-  if (is_n) return PyInt_FromLong(n);
-  return Py_None;
+  int32_t n = self->ptr->get_n_items();
+  return PyInt_FromLong(n);
 }
 
 
@@ -286,9 +306,9 @@ static PyObject *
 py_an_verbose(py_annoy *self, PyObject *args) {
   int verbose;
   if (!self->ptr) 
-    return Py_None;
+    Py_RETURN_NONE;
   if (!PyArg_ParseTuple(args, "i", &verbose))
-    return Py_None;
+    Py_RETURN_NONE;
 
   self->ptr->verbose((bool)verbose);
 
@@ -296,18 +316,34 @@ py_an_verbose(py_annoy *self, PyObject *args) {
 }
 
 
+static PyObject *
+py_an_set_seed(py_annoy *self, PyObject *args) {
+  int q;
+  if (!self->ptr)
+    Py_RETURN_NONE;
+  if (!PyArg_ParseTuple(args, "i", &q))
+    Py_RETURN_NONE;
+
+  self->ptr->set_seed(q);
+
+  Py_RETURN_NONE;
+}
+
+
 static PyMethodDef AnnoyMethods[] = {
-  {"load",	(PyCFunction)py_an_load, METH_VARARGS, ""},
-  {"save",	(PyCFunction)py_an_save, METH_VARARGS, ""},
+  {"load",	(PyCFunction)py_an_load, METH_VARARGS | METH_KEYWORDS, "Loads (mmaps) an index from disk."},
+  {"save",	(PyCFunction)py_an_save, METH_VARARGS | METH_KEYWORDS, "Saves the index to disk."},
   {"get_nns_by_item",(PyCFunction)py_an_get_nns_by_item, METH_VARARGS, ""},
   {"get_nns_by_vector",(PyCFunction)py_an_get_nns_by_vector, METH_VARARGS, ""},
   {"get_item_vector",(PyCFunction)py_an_get_item_vector, METH_VARARGS, ""},
   {"add_item",(PyCFunction)py_an_add_item, METH_VARARGS, ""},
-  {"build",(PyCFunction)py_an_build, METH_VARARGS, ""},
-  {"unload",(PyCFunction)py_an_unload, METH_VARARGS, ""},
+  {"build",(PyCFunction)py_an_build, METH_VARARGS | METH_KEYWORDS, "Builds a forest of `n_trees` trees.\n\nMore trees give higher precision when querying. After calling `build`,\nno more items can be added."},
+  {"unbuild",(PyCFunction)py_an_unbuild, METH_NOARGS, "Unbuilds the tree in order to allows adding new items.\n\nbuild() has to be called again afterwards in order to\nrun queries."},
+  {"unload",(PyCFunction)py_an_unload, METH_NOARGS, "Unloads an index from disk."},
   {"get_distance",(PyCFunction)py_an_get_distance, METH_VARARGS, ""},
-  {"get_n_items",(PyCFunction)py_an_get_n_items, METH_VARARGS, ""},
+  {"get_n_items",(PyCFunction)py_an_get_n_items, METH_NOARGS, "Returns the number of items in the index."},
   {"verbose",(PyCFunction)py_an_verbose, METH_VARARGS, ""},
+  {"set_seed",(PyCFunction)py_an_set_seed, METH_VARARGS, "Sets the seed of Annoy's random number generator."},
   {NULL, NULL, 0, NULL}		 /* Sentinel */
 };
 
